"""Pydantic models."""
from typing import Dict, List, Union

from pydantic import BaseModel, constr
from reasoner_pydantic import Message , Query as ReasonerRequest


class CypherRequest(BaseModel):
    query: str


class SimpleSpecElement(BaseModel):
    source_type: str
    target_type: str
    edge_type: str


SimpleSpecResponse = List[SimpleSpecElement]

TypeSet = constr(regex=r"\w+(:\w+)*")


class TypeSummary(BaseModel):
    nodes_count: int

    class Config:
        schema_extra = {
            "patternProperties": {
                r"\w+(:\w+)*": {
                    "type": "object",
                    "additionalProperties": {
                        "type": "integer"
                    }
                }
            }
        }


# class GraphSummaryResponse(BaseModel):
#     __root__: Dict[TypeSet, TypeSummary]


# print(GraphSummaryResponse.schema_json(indent=4))

GraphSummaryResponse = Dict[TypeSet, TypeSummary]


class CypherDatum(BaseModel):
    row: List
    meta: List


class CypherResult(BaseModel):
    columns: List[str]
    data: List[CypherDatum]


class CypherError(BaseModel):
    code: str
    message: str


class CypherResponse(BaseModel):
    results: List[CypherResult]
    errors: List[CypherError]


PredicatesResponse = Dict[str, Dict[str, List[str]]]



"""FastAPI app."""
import json
import os
import yaml
from typing import Any, Dict, List

from fastapi import Body, Depends, FastAPI
from fastapi.middleware.cors import CORSMiddleware
from fastapi.openapi.utils import get_openapi

from PLATER.services.models import (
    Message, ReasonerRequest, CypherRequest, SimpleSpecResponse, SimpleSpecElement,
    GraphSummaryResponse, CypherResponse, PredicatesResponse
)
from PLATER.services.config import config
from PLATER.services.util.bl_helper import BLHelper
from PLATER.services.util.graph_adapter import GraphInterface
from PLATER.services.util.logutil import LoggingUtil
from PLATER.services.util.overlay import Overlay
from PLATER.services.util.question import Question

TITLE = config.get('PLATER_TITLE', 'Plater API')
VERSION = os.environ.get('PLATER_VERSION', '1.0.0')

logger = LoggingUtil.init_logging(
    __name__,
    config.get('logging_level'),
    config.get('logging_format'),
)

APP = FastAPI()


def get_graph_interface():
    """Get graph interface."""
    return GraphInterface(
        host=config.get('DB_HOST'),
        port=config.get('DB_PORT'),
        auth=(
            config.get('DB_USERNAME'),
            config.get('DB_PASSWORD')
        ),
        db_type=config.get('DB_TYPE'),
        db_name=config.get('DB_NAME')
    )


def get_bl_helper():
    """Get Biolink helper."""
    return BLHelper(config.get('BL_HOST', 'https://bl-lookup-sri.renci.org'))


def get_example(operation: str):
    """Get example for operation."""
    with open(os.path.join(
        os.path.dirname(__file__),
        "..",
        "examples",
        f"{operation}.json",
    )) as stream:
        return json.load(stream)


async def reasoner_api(
        request: ReasonerRequest = Body(
            ...,
            example=get_example("reasoner"),
        ),
        graph_interface: GraphInterface = Depends(get_graph_interface),
) -> Message:
    """Handle TRAPI request."""
    request_json = request.dict()
    question = Question(request_json["message"])
    response = await question.answer(graph_interface)
    request_json.update({'message': response})
    return request_json


APP.add_api_route(
    "/query",
    reasoner_api,
    methods=["POST"],
    response_model=ReasonerRequest,
    summary="Query Reasoner API",
    description="Given a question graph return question graph plus answers.",
    tags=["translator"]
)

APP.add_api_route(
    "/reasonerapi",
    reasoner_api,
    methods=["POST"],
    response_model=ReasonerRequest,
    deprecated=True,
)


async def cypher(
        request: CypherRequest = Body(
            ...,
            example={"query": "MATCH (n) RETURN count(n)"},
        ),
        graph_interface: GraphInterface = Depends(get_graph_interface),
) -> CypherResponse:
    """Handle cypher."""
    request = request.dict()
    results = await graph_interface.run_cypher(
        request["query"],
        return_errors=True,
    )
    return results


APP.add_api_route(
    "/cypher",
    cypher,
    methods=["POST"],
    response_model=CypherResponse,
    summary="Run cypher query",
    description=(
        "Runs cypher query against the Neo4j instance, and returns an "
        "equivalent response expected from a Neo4j HTTP endpoint "
        "(https://neo4j.com/docs/rest-docs/current/)."
    ),
)


async def predicates(
        graph_interface: GraphInterface = Depends(get_graph_interface),
) -> PredicatesResponse:
    """Handle /predicates."""
    return graph_interface.get_schema()


APP.add_api_route(
    "/predicates",
    predicates,
    methods=["GET"],
    response_model=PredicatesResponse,
    summary="Get one-hop connection types",
    description=(
        "Returns an object where outer keys represent source types with "
        "second level keys as targets. And the values of the second level "
        "keys is the type of possible edge typesthat connect these concepts."
    ),
    tags=["translator"]
)


APP.add_api_route(
    "/graph/schema",
    predicates,
    methods=["GET"],
    response_model=Dict,
    summary="Get one-hop connection types",
    description="Get one-hop connection types",
    deprecated=True
)


async def simple_spec(
        source: str = None,
        target: str = None,
        graph_interface: GraphInterface = Depends(get_graph_interface),
        bl_helper: BLHelper = Depends(get_bl_helper),
) -> SimpleSpecResponse:
    """Handle simple spec."""
    source_id = source
    target_id = target
    if source_id or target_id:
        minischema = []
        mini_schema_raw = await graph_interface.get_mini_schema(
            source_id,
            target_id,
        )
        for row in mini_schema_raw:
            source_labels = await bl_helper.get_most_specific_concept(
                row['source_label']
            )
            target_labels = await bl_helper.get_most_specific_concept(
                row['target_label']
            )
            for source_type in source_labels:
                for target_type in target_labels:
                    minischema.append((
                        source_type,
                        row['predicate'],
                        target_type,
                    ))
        minischema = list(set(minischema))  # remove dups
        return list(map(lambda x: SimpleSpecElement(**{
                'source_type': x[0],
                'target_type': x[2],
                'edge_type': x[1],
            }), minischema))
    else:
        schema = graph_interface.get_schema()
        reformatted_schema = []
        for source_type in schema:
            for target_type in schema[source_type]:
                for edge in schema[source_type][target_type]:
                    reformatted_schema.append(SimpleSpecElement(**{
                        'source_type': source_type,
                        'target_type': target_type,
                        'edge_type': edge
                    }))
        return reformatted_schema


APP.add_api_route(
    "/simple_spec",
    simple_spec,
    methods=["GET"],
    response_model=SimpleSpecResponse,
    summary="Get one-hop connection schema",
    description=(
        "Returns a list of available predicates when choosing a single source "
        "or target curie. Calling this endpoint with no query parameters will "
        "return all possible hops for all types."
    ),
)


async def get_reasoner_api(
        graph_interface: GraphInterface = Depends(get_graph_interface),
) -> List[Message]:
    """Handle GET /reasonerapi."""
    return Question.transform_schema_to_question_template(
        graph_interface.get_schema()
    )


APP.add_api_route(
    "/reasonerapi",
    get_reasoner_api,
    methods=["GET"],
    response_model=List[Message],
    summary="Get question templates",
    description=(
        "Returns a list of question templates that can be used to query this "
        "plater instance."
    ),
)


async def graph_summary(
        graph_interface: GraphInterface = Depends(get_graph_interface),
) -> GraphSummaryResponse:
    """Handle GET /graph_summary."""
    graph_interface.get_schema()
    return graph_interface.summary


APP.add_api_route(
    "/graph/summary",
    graph_summary,
    methods=["GET"],
    response_model=GraphSummaryResponse,
    summary="Get graph summary",
    description="Returns a summary of the graph.",
)


async def overlay(
        request: ReasonerRequest = Body(
            ...,
            example={"message": get_example("overlay")},
        ),
        graph_interface: GraphInterface = Depends(get_graph_interface),
) -> Message:
    """Handle TRAPI request."""
    overlay_class = Overlay(graph_interface)
    return await overlay_class.overlay_support_edges(request.dict()["message"])


APP.add_api_route(
    "/overlay",
    overlay,
    methods=["POST"],
    response_model=Message,
    description=(
        "Given a ReasonerAPI graph, add support edges for any nodes linked in "
        "result bindings."
    ),
    summary="Overlay results with available connections between each node.",
    tags=["translator"]
)


async def about() -> Any:
    """Handle /about."""
    with open('about.json') as f:
        return json.load(f)


APP.add_api_route(
    "/about",
    about,
    methods=["GET"],
    response_model=Any,
    summary="JSON about dataset",
    description="Returns a JSON describing dataset.",
)


async def one_hop(
        source_type: str,
        target_type: str,
        curie: str,
        graph_interface: GraphInterface = Depends(get_graph_interface),
) -> List[Dict]:
    """Handle one-hop."""
    return await graph_interface.get_single_hops(
        source_type,
        target_type,
        curie,
    )


APP.add_api_route(
    "/{source_type}/{target_type}/{curie}",
    one_hop,
    methods=["GET"],
    response_model=List,
    summary=(
        "Get one hop results from source type to target type. "
        "Note: Please GET /predicates to determine what target goes "
        "with a source"
    ),
    description=(
        "Returns one hop paths from `source_node_type`  with `curie` "
        "to `target_node_type`."
    ),
)


async def node(
        node_type: str,
        curie: str,
        graph_interface: GraphInterface = Depends(get_graph_interface),
) -> List[List[Dict]]:
    """Handle node lookup."""
    return await graph_interface.get_node(
        node_type,
        curie,
    )


APP.add_api_route(
    "/{node_type}/{curie}",
    node,
    methods=["GET"],
    response_model=List,
    summary="Find `node` by `curie`",
    description="Returns `node` matching `curie`.",
)


def construct_open_api_schema():

    if APP.openapi_schema:
        return APP.openapi_schema
    open_api_schema = get_openapi(
        title=TITLE,
        version=VERSION,
        description='',
        routes=APP.routes
    )

    open_api_extended_file_path = config.get_resource_path('../openapi-config.yaml')
    with open(open_api_extended_file_path) as open_api_file:
        open_api_extended_spec = yaml.load(open_api_file, Loader=yaml.SafeLoader)

    x_translator_extension = open_api_extended_spec.get("x-translator")
    contact_config = open_api_extended_spec.get("contact")
    terms_of_service = open_api_extended_spec.get("termsOfService")
    servers_conf = open_api_extended_spec.get("servers")
    tags = open_api_extended_spec.get("tags")
    title_override = open_api_extended_spec.get("title") or TITLE
    description = open_api_extended_spec.get("description")

    if tags:
        open_api_schema['tags'] = tags

    if x_translator_extension:
        # if x_translator_team is defined amends schema with x_translator extension
        open_api_schema["info"]["x-translator"] = x_translator_extension

    if contact_config:
        open_api_schema["info"]["contact"] = contact_config

    if terms_of_service:
        open_api_schema["info"]["termsOfService"] = terms_of_service

    if description:
        open_api_schema["info"]["description"] = description

    if title_override:
        open_api_schema["info"]["title"] = title_override

    if servers_conf:
        open_api_schema["servers"] = servers_conf

    return open_api_schema


APP.openapi_schema = construct_open_api_schema()

# CORS
APP.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

if __name__=="__main__":
    import uvicorn

    uvicorn.run(
        port = 8080,
        app = APP
    )



import asyncio
from functools import reduce
import httpx
from PLATER.services.config import config


class BLHelper:
    def __init__(self, bl_url=config.get('bl_url')):
        self.bl_url = bl_url

    @staticmethod
    async def make_request(url):
        async with httpx.AsyncClient() as session:
            response = await session.get(url)
            if response.status_code == 200:
                return response.json()
            else:
                return None

    async def get_most_specific_concept(self, concept_list: list) -> list:
        """
        Given a list of concepts find the most specific set of concepts.
        """
        tasks = []
        for concept in concept_list:
            parent_url = f"{self.bl_url}/bl/{concept}/ancestors"
            tasks.append(BLHelper.make_request(parent_url))
        response = await asyncio.gather(*tasks, return_exceptions=False)
        parents = list(reduce(lambda acc, value: acc + value, filter(lambda x: x, response), []))
        return list(filter(lambda x: x not in parents, concept_list))



import base64
import traceback
import httpx
from PLATER.services.config import config
from PLATER.services.util.logutil import LoggingUtil
from reasoner.cypher import get_query

logger = LoggingUtil.init_logging(__name__,
                                  config.get('logging_level'),
                                  config.get('logging_format')
                                  )


class Neo4jHTTPDriver:
    def __init__(self, host: str, port: int,  auth: set, scheme: str = 'http'):
        self._host = host
        self._neo4j_transaction_endpoint = "/db/data/transaction/commit"
        self._scheme = scheme
        self._full_transaction_path = f"{self._scheme}://{self._host}:{port}{self._neo4j_transaction_endpoint}"
        self._port = port
        self._supports_apoc = None
        self._header = {
                'Accept': 'application/json; charset=UTF-8',
                'Content-Type': 'application/json',
                'Authorization': 'Basic %s' % base64.b64encode(f"{auth[0]}:{auth[1]}".encode('utf-8')).decode('utf-8')
            }
        # ping and raise error if neo4j doesn't respond.
        logger.debug('PINGING NEO4J')
        self.ping()
        logger.debug('CHECKING IF NEO4J SUPPORTS APOC')
        self.check_apoc_support()
        logger.debug(f'SUPPORTS APOC : {self._supports_apoc}')

    async def post_request_json(self, payload):
        async with httpx.AsyncClient(timeout=600) as session:
            response = await session.post(self._full_transaction_path, json=payload, headers=self._header)
            if response.status_code != 200:
                logger.error(f"[x] Problem contacting Neo4j server {self._host}:{self._port} -- {response.status_code}")
                txt = response.text
                logger.debug(f"[x] Server responded with {txt}")
            else:
                return response.json()

    def ping(self):
        """
        Pings the neo4j backend.
        :return:
        """
        neo4j_db_labels_endpoint = "/db/data/labels"
        ping_url = f"{self._scheme}://{self._host}:{self._port}{neo4j_db_labels_endpoint}"
        logger.info(ping_url)
        # if we can't contact neo4j, we should exit.
        try:
            import time
            now = time.time()
            response = httpx.get(ping_url, headers=self._header)
            later = time.time()
            time_taken = later - now
            logger.debug(f'Contacting neo4j took {time_taken} seconds.')
            if time_taken > 5:  # greater than 5 seconds it's not healthy
                logger.warn(f"Contacting neo4j took more than 5 seconds ({time_taken}). Neo4j might be stressed.")
            if response.status_code != 200:
                raise Exception(f'server returned {response.status_code}')
        except Exception as e:
            logger.error(f"Error contacting Neo4j @ {ping_url} -- Exception raised -- {e}")
            logger.debug(traceback.print_exc())
            raise RuntimeError('Connection to Neo4j could not be established.')

    async def run(self, query, return_errors=False):
        """
        Runs a neo4j query async.
        :param query: Cypher query.
        :type query: str
        :return: result of query.
        :rtype: dict
        """
        # make the statement dictionary
        payload = {
            "statements": [
                {
                    "statement": f"{query}"
                }
            ]
        }

        response = await self.post_request_json(payload)
        errors = response.get('errors')
        if errors:
            if return_errors:
                return response
            logger.error(f'Neo4j returned `{errors}` for cypher {query}.')
            raise RuntimeWarning(f'Error running cypher {query}.')
        return response

    def run_sync(self, query):
        """
        Runs a neo4j query. Can cause the async loop to block.
        :param query:
        :return:
        """
        payload = {
            "statements": [
                {
                    "statement": f"{query}"
                }
            ]
        }
        response = httpx.post(
            self._full_transaction_path,
            headers=self._header,
            timeout=600,
            json=payload).json()
        errors = response.get('errors')
        if errors:
            logger.error(f'Neo4j returned `{errors}` for cypher {query}.')
            raise RuntimeWarning(f'Error running cypher {query}.')
        return response

    def convert_to_dict(self, response: dict) -> list:
        """
        Converts a neo4j result to a structured result.
        :param response: neo4j http raw result.
        :type response: dict
        :return: reformatted dict
        :rtype: dict
        """
        results = response.get('results')
        array = []
        if results:
            for result in results:
                cols = result.get('columns')
                if cols:
                    data_items = result.get('data')
                    for item in data_items:
                        new_row = {}
                        row = item.get('row')
                        for col_name, col_value in zip(cols, row):
                            new_row[col_name] = col_value
                        array.append(new_row)
        return array

    def check_apoc_support(self):
        apoc_version_query = 'call apoc.help("meta")'
        if self._supports_apoc is None:
            try:
                self.run_sync(apoc_version_query)
                self._supports_apoc = True
            except:
                self._supports_apoc = False
        return self._supports_apoc

    async def answer_TRAPI_question(self, trapi_question):
        cypher_query = get_query(trapi_question)
        logger.info("RUNNING TRAPI QUERY:")
        logger.info(cypher_query)
        results = await self.run(cypher_query)
        results_dict = self.convert_to_dict(results)[0]
        results_dict.update({'query_graph': trapi_question})
        return results_dict





import logging
from logging.handlers import RotatingFileHandler
import os

class LoggingUtil(object):
    """ Logging utility controlling format and setting initial logging level """

    @staticmethod
    def init_logging(name, level=logging.INFO, format_sel='medium', log_file_level=None):
        log_file_path = os.path.join(os.path.dirname(__file__), '../../logs/plater.log')
        # get a logger
        logger = logging.getLogger(__name__)

        # returns a new logger if its not the root
        if not logger.parent.name == 'root':
            return logger

        # define the output types
        format_types = {
            "short": '[%(name)s.%(funcName)s] : %(message)s',
            "medium": '[%(name)s.%(funcName)s] - %(asctime)-15s: %(message)s',
            "long": '[%(name)s.%(funcName)s] - %(asctime)-15s %(filename)s %(levelname)s: %(message)s'
        }[format_sel]

        # create a stream handler (default to console)
        stream_handler = logging.StreamHandler()

        # create a formatter
        formatter = logging.Formatter(format_types)

        # set the formatter on the console stream
        stream_handler.setFormatter(formatter)

        # get the name of this logger
        logger = logging.getLogger(name)

        # set the logging level
        logger.setLevel(level)

        # if there was a file path passed in use it
        if log_file_path is not None:
            # create a rotating file handler, 100mb max per file with a max number of 10 files
            file_handler = RotatingFileHandler(filename=log_file_path, maxBytes=1000000, backupCount=10)

            # set the formatter
            file_handler.setFormatter(formatter)

            # if a log level for the file was passed in use it
            if log_file_level is not None:
                level = log_file_level

            # set the log level
            file_handler.setLevel(level)

            # add the handler to the logger
            logger.addHandler(file_handler)

        # add the console handler to the logger
        logger.addHandler(stream_handler)

        # return to the caller
        return logger



import copy
from functools import reduce
from PLATER.services.util.graph_adapter import GraphInterface
import time
from reasoner.cypher import get_query

class Question:

    #SPEC VARS
    QUERY_GRAPH_KEY='query_graph'
    KG_ID_KEY='id'
    QG_ID_KEY='id'
    ANSWERS_KEY='results'
    KNOWLEDGE_GRAPH_KEY='knowledge_graph'
    NODES_LIST_KEY='nodes'
    EDGES_LIST_KEY='edges'
    NODE_TYPE_KEY='category'
    EDGE_TYPE_KEY='predicate'
    SOURCE_KEY='subject'
    TARGET_KEY='object'
    NODE_BINDINGS_KEY='node_bindings'
    EDGE_BINDINGS_KEY='edge_bindings'
    CURIE_KEY = 'curie'

    def __init__(self, question_json):
        self._question_json = copy.deepcopy(question_json)

    def compile_cypher(self):
        return get_query(self._question_json[Question.QUERY_GRAPH_KEY])

    async def answer(self, graph_interface: GraphInterface):
        """
        Updates the query graph with answers from the neo4j backend
        :param graph_interface: interface for neo4j
        :return: None
        """
        s = time.time()
        answer = await graph_interface.answer_trapi_question(self._question_json['query_graph'])
        end = time.time()
        print(f'grabbing results took {end - s}')
        return answer

    @staticmethod
    def transform_schema_to_question_template(graph_schema):
        """
        Returns array of Templates given a graph schema
        Eg: if schema looks like
           {
            "Type 1" : {
                "Type 2": [
                    "edge 1"
                ]
            }
           }
           We would get
           {
            "question_graph": {
                "nodes" : {
                    "n1": {
                        "id": "{{ curie }}",
                        "category": "Type 1"
                    },
                    "n2": {
                        "id" : "{{ curie }}",
                        "category": "Type 2"
                    }
                },
                "edges":{
                    "e1": {
                        "predicate": "edge 1",
                        "subject": "n1",
                        "object": "n2"
                    }
                ]
            }
           }
        :param graph_schema:
        :return:
        """
        question_templates = []
        for source_type in graph_schema:
            target_set = graph_schema[source_type]
            for target_type in target_set:
                question_graph = {
                    Question.NODES_LIST_KEY: {
                        "n1": {
                            'id': None,
                            Question.NODE_TYPE_KEY: source_type,
                        },
                        "n2": {
                            'id': None,
                            Question.NODE_TYPE_KEY: target_type,
                        }
                    },
                    Question.EDGES_LIST_KEY: []
                }
                edge_set = target_set[target_type]
                question_graph[Question.EDGES_LIST_KEY] = {}
                for index, edge_type in enumerate(set(edge_set)):
                    edge_dict = {
                        Question.SOURCE_KEY: "n1",
                        Question.TARGET_KEY: "n2",
                        Question.EDGE_TYPE_KEY: edge_type
                    }
                    question_graph[Question.EDGES_LIST_KEY][f"e{index}"] = edge_dict
            question_templates.append({Question.QUERY_GRAPH_KEY: question_graph})
        return question_templates




#from distutils.core import setup
from setuptools import setup
import os
# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()
    long_description = f"See the Homepage for a better formatted version.\n {long_description}"
setup(
    name = 'PLATER-GRAPH',
    packages = ['PLATER',
    'PLATER/services',
    'PLATER/tests',
    'PLATER/services/util',
    'PLATER/services/util/drivers'],
    data_files= [('',['PLATER/services/plater.conf', 'PLATER/logs/plater.log'])],
    version = '1.6',
    description = 'TranslatorAPI Interface for graph databases.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author = '',
    author_email = 'yaphetkg@renci.org',
    install_requires = [
      'coverage',
      'pyaml==20.3.1',
      'pytest==5.4.1',
      'pytest-asyncio==0.14.0',
      'starlette==0.13.6',
      'uvicorn==0.11.7',
      'httpx',
      'redis'
    ],
    include_package_data=True,
    entry_points = {
    },
    dependencies= [
    'git+git://github.com/patrickkwang/biolink-model-toolkit@master#egg=bmt',
    'git+https://github.com/ranking-agent/reasoner.git',
    'git+https://github.com/TranslatorSRI/reasoner-pydantic@v1.0#egg=reasoner-pydantic',
    'git+https://github.com/patrickkwang/fastapi#egg=fastapi',
    'git+https://github.com/redislabs/redisgraph-py.git'
    ],
    url = 'http://github.com/YaphetKG/plater.git',
    download_url = 'http://github.com/yaphetkg/plater/archive/GRAPH-PLATER-1.6.tar.gz',
    keywords = [ 'knowledge', 'network', 'graph', 'biomedical' ],
    classifiers = [ ],
)
